"""
Code relating to the command line interface to acclimatise
"""
import sys
from pathlib import Path
from typing import Iterable, Tuple

import click

from acclimatise import WrapperGenerator, best_cmd, explore_command, parse_help
from acclimatise.flag_parser.parser import CliParser

# Some common options
opt_pos = click.option(
    "--pos/--no-pos",
    help=(
        "Include (default) or don't include positional arguments, for example because the help formatting has some "
        "misleading sections that look like positional arguments"
    ),
)
opt_generate_names = click.option(
    "--generate-names",
    "-g",
    is_flag=True,
    help=(
        "Rather than using the long flag to generate the argument name, generate them automatically using the "
        "flag description. Generally helpful if there are no long flags, only short flags."
    ),
)
opt_case = click.option(
    "--case",
    "-c",
    type=click.Choice(WrapperGenerator.cases),
    help=(
        "Which case to use for variable names. If not set, defaults to the language defaults: snake_case for CWL"
        " and snake_case for WDL"
    ),
    default="snake",
)
opt_cmd = click.argument("cmd", nargs=-1)


@click.group()
def main():
    pass


@main.command(help="Run an executable and explore all subcommands")
@opt_cmd
@opt_pos
@opt_case
@opt_generate_names
@click.option(
    "--format",
    "-f",
    "formats",
    type=click.Choice(["wdl", "cwl", "yml"]),
    multiple=True,
    default=("yml", "wdl", "cwl"),
    help="The language in which to output the CLI wrapper",
)
@click.option(
    "--out-dir",
    "-o",
    type=Path,
    help="Directory in which to put the output files",
    required=True,
)
@click.option(
    "--help-flag",
    "-f",
    type=str,
    help="Flag to append to the end of the command to make it output help text",
)
@click.option(
    "--subcommands/--no-subcommands", default=True, help="Look for subcommands"
)
def explore(
    cmd: Iterable[str],
    out_dir: Path,
    formats: Tuple[str],
    subcommands: bool,
    case: str,
    generate_names: bool,
    pos: bool,
    help_flag: str,
):
    # Optionally parse subcommands
    kwargs = {}
    if help_flag is not None:
        kwargs = {"flags": ([help_flag],)}
    if subcommands:
        command = explore_command(list(cmd), **kwargs)
    else:
        command = best_cmd(list(cmd), **kwargs)

    for format in formats:
        converter_cls = WrapperGenerator.choose_converter(format)
        converter = converter_cls(
            generate_names=generate_names, ignore_positionals=not pos, case=case,
        )
        list(converter.generate_tree(command, out_dir))


@main.command(
    help="Read a command help from stdin and output a tool definition to stdout"
)
@opt_cmd
@opt_pos
@opt_generate_names
@opt_case
@click.option(
    "--format",
    "-f",
    type=click.Choice(["wdl", "cwl", "yml"]),
    default="cwl",
    help="The language in which to output the CLI wrapper",
)
def pipe(cmd, pos, generate_names, case, format):
    stdin = "".join(sys.stdin.readlines())
    command = parse_help(cmd, stdin)

    converter_cls = WrapperGenerator.choose_converter(format)
    converter = converter_cls(
        generate_names=generate_names, ignore_positionals=not pos, case=case,
    )
    output = converter.generate_wrapper(command)
    print(output)


@main.command(help="Output a representation of the internal grammar")
@opt_pos
def grammar(pos):
    parser = CliParser(pos)
    print(str(parser.flags))


if __name__ == "__main__":
    main()



from dataclasses import dataclass


@dataclass
class UsageElement:
    text: str
    """
    The name of this element, as defined in the usage section
    """

    optional: bool = False
    """
    Whether or not this element is required
    """

    variable: bool = False
    """
    True if this is a variable, ie you are supposed to replace this text with your own, False if this is a constant
    that you shouldn't change, e.g. the name of the application
    """

    # flag: bool = False
    """
    True if this is a flag (starts with dashes) and not a regular element
    """

    repeatable: bool = False
    """
    If this flag/argument can be used multiple times
    """



"""
Contains the CLI data model
"""
import abc
import enum
import itertools
import re
import typing
import unicodedata
from abc import abstractmethod

import spacy
import wordsegment
from dataclasses import InitVar, dataclass, field
from ruamel.yaml import YAML, yaml_object
from spacy import tokens

from acclimatise import cli_types
from acclimatise.yaml import yaml


@yaml_object(yaml)
@dataclass
class CliArgument:
    """
    A generic parent class for both named and positional CLI arguments
    """

    #: Description of the function of this argument
    description: str

    #: Whether the existence of this argument is supported by it appearing in the usage
    usage_supported: bool = field(default=False, init=False)

    @staticmethod
    def tokens_to_name(tokens: typing.List[tokens.Token]):
        return re.sub("[^\w]", "", "".join([tok.text.capitalize() for tok in tokens]))

    @abstractmethod
    def full_name(self) -> str:
        pass

    @abstractmethod
    def get_type(self) -> cli_types.CliType:
        """
        Return a type object indicating the type of data this argument holds. e.g. If it's an array type this will be a CliList.
        """
        pass

    def name_to_words(self) -> typing.Iterable[str]:
        """
        Splits this argument's name into multiple words
        """
        # Load wordsegment the first time
        if len(wordsegment.WORDS) == 0:
            wordsegment.load()

        base = self.full_name().lstrip("-")

        # Replace symbols with their unicode names
        translated = re.sub(
            "[\W\S\-_]", lambda symbol: unicodedata.name(symbol[0]).lower(), base
        )

        dash_tokens = re.split("[-_ ]", translated)
        segment_tokens = itertools.chain.from_iterable(
            [wordsegment.segment(w) for w in dash_tokens]
        )
        return segment_tokens

    def name_to_camel(self) -> str:
        """
        Gets a representation of this argument in CamelCase
        """
        words = list(self.name_to_words())
        cased = [words[0].lower()] + [segment.capitalize() for segment in words]
        return "".join(cased)

    def name_to_snake(self) -> str:
        """
        Gets a representation of this argument in snake case
        """
        words = self.name_to_words()
        return "_".join([segment for segment in words])

    def generate_name(self) -> str:
        """
        Generate a 1-3 word variable name for this flag, by parsing the description
        """
        try:
            nlp = spacy.load("en")
        except IOError:
            nlp = None

        if nlp is None:
            raise Exception(
                "Spacy model doesn't exist! Install it with `python -m spacy download en`"
            )

        no_brackets = re.sub("[\[({].+[\])}]", "", self.description)
        words = nlp(no_brackets)
        # for chunk in words.noun_chunks:
        #     if chunk.root.dep_ == 'ROOT':
        #         return chunk.text
        root = None

        # Find the actual root
        for word in words:
            if (word.dep_ == "ROOT" or word.dep == "nsubj") and word.pos_ == "NOUN":
                root = word

        # If that isn't there, get the first noun
        if root is None:
            for word in words:
                if word.pos_ == "NOUN":
                    root = word

        # If that isn't there, get the first word
        if root is None:
            return self.tokens_to_name(words)

        subtree = list(root.subtree)
        if len(subtree) < 4:
            # If the whole subtree is a reasonable length, use that
            return self.tokens_to_name(subtree)
        else:
            good_children = [tok for tok in subtree if tok.dep_ != "prep"]
            if len(good_children) >= 1:
                subtree = sorted([root, good_children[0]], key=lambda tok: tok.i)
                # Otherwise, just add the first child token
                return self.tokens_to_name(subtree)
            else:
                return self.tokens_to_name([root])


@yaml_object(yaml)
@dataclass
class Command:
    """
    Class representing an entire command or subcommand, e.g. `bwa mem` or `grep`
    """

    def __post_init__(self):
        # Store certain special flags in their own fields
        if self.help_flag is None:
            for flag in self.named:
                if (
                    "--help" in flag.synonyms
                    or "-help" in flag.synonyms
                    or ("-h" in flag.synonyms and isinstance(flag.args, EmptyFlagArg))
                ):
                    self.help_flag = flag
                    self.named.remove(flag)

        if self.version_flag is None:
            for flag in self.named:
                if "--version" in flag.synonyms:
                    self.version_flag = flag
                    self.named.remove(flag)

        if self.usage_flag is None:
            for flag in self.named:
                if "--usage" in flag.synonyms:
                    self.usage_flag = flag
                    self.named.remove(flag)

    @property
    def as_filename(self) -> str:
        """
        Returns a sample filename that might be used to store this command (without a suffix)
        """
        return "_".join(self.command).replace("-", "_")

    def command_tree(self) -> typing.Generator["Command", None, None]:
        """
        Returns a generator over the entire command tree. e.g. if this command has 2 subcommands, each with 2
            subcommands, this will return a generator with 7 Commands
        """
        yield self
        for command in self.subcommands:
            yield from command.command_tree()

    #
    # def __init__(
    #     self,
    #     command: typing.List[str],
    #     positional: typing.List["Positional"],
    #     named: typing.List["Flag"],
    #     **kwargs
    # ):
    #     super().__init__(**kwargs)
    #
    #     self.command = command
    #     self.named = []
    #
    #     # Put the help and usage flag into separate variables
    #     for flag in named:
    #         if (
    #             "--help" in flag.synonyms
    #             or "-help" in flag.synonyms
    #             or ("-h" in flag.synonyms and isinstance(flag.args, EmptyFlagArg))
    #         ):
    #             self.help_flag = flag
    #         elif "--usage" in flag.synonyms:
    #             self.usage_flag = flag
    #         elif "--version" in flag.synonyms:  # or "-v" in flag.synonyms:
    #             self.version_flag = flag
    #         else:
    #             self.named.append(flag)
    #     self.positional = positional

    positional: typing.List["Positional"]
    named: typing.List["Flag"]
    command: typing.List[str]

    subcommands: typing.List["Command"] = field(default_factory=list)
    help_flag: typing.Optional["Flag"] = None
    usage_flag: typing.Optional["Flag"] = None
    version_flag: typing.Optional["Flag"] = None


@yaml_object(yaml)
@dataclass
class Positional(CliArgument):
    """
    A positional command-line argument. This probably means that it is required, and has no arguments like flags do
    """

    def full_name(self) -> str:
        """
        Getting the full name for a positional argument is easy - it's just the parameter name
        """
        return self.name

    position: int
    name: str
    description: str
    optional: bool = False

    def get_type(self) -> cli_types.CliType:
        # Try the the flag name, then the description in that order

        name_type = infer_type(self.name)
        if name_type is not None:
            return name_type

        flag_type = infer_type(self.full_name())
        if flag_type is not None:
            return flag_type

        return cli_types.CliString()


@yaml_object(yaml)
@dataclass
class Flag(CliArgument):
    """
    Represents one single flag, with all synonyms for it, and all arguments, e.g. `-h, --help`
    """

    synonyms: typing.List[str]
    description: typing.Optional[str]
    args: "FlagArg"
    optional: bool = True

    def get_type(self) -> cli_types.CliType:
        # Try the argument name, then the flag name, then the description in that order
        arg_type = self.args.get_type()
        if arg_type is not None:
            return arg_type

        flag_type = infer_type(self.full_name())
        if flag_type is not None:
            return flag_type

        description_type = infer_type(self.description)
        if description_type is not None:
            return description_type

        return cli_types.CliString()

    def full_name(self) -> str:
        """
        Getting the full name for a named flag is slightly harder, we need to find the longest synonym
        """
        return self.longest_synonym

    @staticmethod
    def from_synonyms(
        synonyms: typing.Iterable["FlagSynonym"], description: typing.Optional[str]
    ):
        """
        Creates a usable Flag object by combining the synonyms provided
        """
        synonym_str = []
        args = None
        arg_count = float("-inf")

        for synonym in synonyms:
            synonym_str.append(synonym.name)
            if synonym.argtype.num_args() > arg_count:
                arg_count = synonym.argtype.num_args()
                args = synonym.argtype

        return Flag(synonyms=synonym_str, args=args, description=description)

    @property
    def longest_synonym(self) -> str:
        """
        Returns the longest synonym this flag has. e.g. for `-h, --help`, it will return `--help`
        """
        return max(self.synonyms, key=lambda synonym: len(synonym))

    @property
    def shortest_synonym(self) -> str:
        """
        Returns the shortest synonym this flag has. e.g. for `-h, --help`, it will return `-h`
        """
        return min(self.synonyms, key=lambda synonym: len(synonym))


@yaml_object(yaml)
@dataclass
class FlagSynonym:
    """
    Internal class for storing the arguments for a single synonym
    """

    name: str
    """
    The entire flag string, e.g. "-n" or "--lines"
    """

    argtype: "FlagArg"
    """
    The number and type of arguments that this flag takes
    """

    @property
    def capital(self):
        return "".join(
            [
                segment.capitalize()
                for segment in re.split("[-_]", self.name.lstrip("-"))
            ]
        )


int_re = re.compile("(int(eger)?)|size|length|max|min", flags=re.IGNORECASE)
str_re = re.compile("str(ing)?", flags=re.IGNORECASE)
float_re = re.compile("float|decimal", flags=re.IGNORECASE)
bool_re = re.compile("bool(ean)?", flags=re.IGNORECASE)
file_re = re.compile("file|path", flags=re.IGNORECASE)
dir_re = re.compile("folder|directory", flags=re.IGNORECASE)


def infer_type(string) -> typing.Optional[cli_types.CliType]:
    """
    Reads a string (argument description etc) to find hints about what type this argument might be. This is
    generally called by the get_type() methods
    """
    if bool_re.match(string):
        return cli_types.CliBoolean()
    elif float_re.match(string):
        return cli_types.CliFloat()
    elif int_re.match(string):
        return cli_types.CliInteger()
    elif file_re.match(string):
        return cli_types.CliFile()
    elif dir_re.match(string):
        return cli_types.CliDir()
    elif str_re.match(string):
        return cli_types.CliString()
    else:
        return cli_types.CliString()


@yaml_object(yaml)
@dataclass
class FlagArg(abc.ABC):
    """
    The data model for the argument or arguments for a flag, for example a flag might have no arguments, it might have
    one argument, it might accept one option from a list of options, or it might accept an arbitrary number of inputs
    """

    @abc.abstractmethod
    def get_type(self) -> cli_types.CliType:
        """
        Return a type object indicating the type of data this argument holds. e.g. If it's an array type this will be a
        CliList.
        """
        pass

    @abc.abstractmethod
    def num_args(self) -> int:
        """
        Calculate the multiplicity of this argument
        """
        pass


@yaml_object(yaml)
@dataclass
class EmptyFlagArg(FlagArg):
    """
    A flag that has no arguments, e.g. `--quiet` that is either present or not present
    """

    def num_args(self) -> int:
        return 0

    def get_type(self):
        return cli_types.CliBoolean()


@yaml_object(yaml)
@dataclass
class OptionalFlagArg(FlagArg):
    """
    When the flag has multiple arguments, some of which are optional, e.g.
    -I FLOAT[,FLOAT[,INT[,INT]]]
    """

    names: list
    separator: str

    def num_args(self) -> int:
        return len(self.names)

    def get_type(self):
        return cli_types.CliTuple([infer_type(arg) for arg in self.names])


@yaml_object(yaml)
@dataclass
class SimpleFlagArg(FlagArg):
    """
    When a flag has one single argument, e.g. `-e PATTERN`, where PATTERN is the argument
    """

    name: str

    def num_args(self) -> int:
        return 1

    def get_type(self):
        return infer_type(self.name)


@yaml_object(yaml)
@dataclass
class RepeatFlagArg(FlagArg):
    """
    When a flag accepts 1 or more arguments, e.g. `--samout SAMOUTS [SAMOUTS ...]`
    """

    name: str

    def num_args(self) -> int:
        return 1

    def get_type(self):
        t = infer_type(self.name)
        return cli_types.CliList(t)


@yaml_object(yaml)
@dataclass
class ChoiceFlagArg(FlagArg):
    """
    When a flag accepts one option from a list of options, e.g. `-s {yes,no,reverse}`
    """

    choices: typing.Set[str]

    def get_type(self):
        e = enum.Enum(
            "".join([choice.capitalize() for choice in self.choices]),
            list(self.choices),
        )
        return cli_types.CliEnum(e)

    def num_args(self) -> int:
        return 1



from acclimatise.flag_parser.elements import *


def pick(*args):
    def action(s, loc, toks):
        if len(args) == 1:
            return toks[args[0]]
        else:
            return [toks[arg] for arg in args]

    return action


class CliParser:
    def parse_command(self, cmd, name) -> Command:
        all_flags = list(itertools.chain.from_iterable(self.flags.searchString(cmd)))
        named = [flag for flag in all_flags if isinstance(flag, Flag)]
        positional = [flag for flag in all_flags if isinstance(flag, Positional)]
        return Command(command=name, positional=positional, named=named)

    def __init__(self, parse_positionals=True):
        stack = [1]

        self.indented_desc = customIndentedBlock(
            desc_line, indentStack=stack, indent=True, terminal=True
        ).setParseAction(lambda s, loc, toks: " ".join([tok[0] for tok in toks[0]]))

        self.description = self.indented_desc.copy().setName(
            "description"
        )  # Optional(one_line_desc) + Optional(indented_desc)
        # A description that takes up one line
        # one_line_desc = SkipTo(LineEnd())

        # A flag description that makes up an indented block
        # originalTextFor(SkipTo(flag_prefix ^ LineEnd()))

        # The entire flag documentation, including all synonyms and description
        self.flag = (
            (flag_synonyms + self.description)
            .setName("flag")
            .setParseAction(
                lambda s, loc, toks: (
                    Flag.from_synonyms(synonyms=toks[0:-1], description=toks[-1])
                )
            )
        )

        self.positional = (
            # Unlike with flags, we have to be a bit pickier about what defines a positional because it's very easy
            # for a paragraph of regular text to be parsed as a positional. So we add a minimum of 2 spaces separation
            (cli_id + White(min=2).suppress() + self.description)
            .setName("positional")
            .setParseAction(
                lambda s, loc, toks: Positional(
                    name=toks[0], description=toks[1], position=-1
                )
            )
        )
        """
        A positional argument, e.g. a required argument that doesn't use dashes
        """

        def visit_flags(s, loc, toks):
            # Give the correct position to the positional arguments
            processed = []
            counter = 0
            flags = toks[0]

            for (flag,) in flags:
                if isinstance(flag, Positional):
                    flag.position = counter
                    counter += 1
                processed.append(flag)

            return processed

        def visit_colon_block(s, loc, toks):
            toks = visit_flags(s, loc, toks)
            if len(toks) < 2:
                raise ParseException(
                    "If the block starts with a colon it must match two or more tokens",
                    loc=loc,
                )
            else:
                return toks

        if parse_positionals:
            block_element = self.flag ^ self.positional
        else:
            block_element = self.flag

        self.flag_block = customIndentedBlock(
            block_element, indentStack=stack, indent=True, lax=True
        ).setName("flag_block")

        # self.flag_block.skipWhitespace = True

        self.colon_block = Literal(
            ":"
        ).suppress() + self.flag_block.copy().setParseAction(visit_colon_block)

        self.newline_block = (
            LineStart().leaveWhitespace()
            + White().suppress()
            + self.flag_block.copy().setParseAction(visit_flags)
        )

        # self.colon_block.skipWhitespace = True
        # self.newline_block.skipWhitespace = True

        # A flag block can start with a colon, but then it must have 2 or more flags. If it starts with a newline it
        # only has to have one flag at least
        self.flags = self.newline_block ^ self.colon_block  # .leaveWhitespace()
        self.flags.skipWhitespace = False

        self.flag_section_header = Regex("(arguments|options):", flags=re.IGNORECASE)
        self.flag_section = (self.flag_section_header + self.flags).setParseAction(
            lambda s, loc, toks: toks[1:]
        )



"""
Contains the objects that represent a "type" of data a flag argument might store
"""
import functools
import typing
from enum import Enum

from dataclasses import dataclass


@dataclass(unsafe_hash=True)
class CliType:
    @property
    def representable(self) -> set:
        """
        Returns a set of types that this type could alternatively be represented as.
        Adds the class's own type to the _representable set
        """
        return self._representable.union({type(self)})

    # The list of types that this specific type could be representable as
    _representable = set()


@dataclass(unsafe_hash=True)
class CliEnum(CliType):
    enum: Enum


@dataclass(unsafe_hash=True)
class CliFloat(CliType):
    pass


@dataclass(unsafe_hash=True)
class CliInteger(CliType):
    _representable = {CliFloat}


@dataclass(unsafe_hash=True)
class CliString(CliType):
    pass


@dataclass(unsafe_hash=True)
class CliBoolean(CliType):
    pass


@dataclass(unsafe_hash=True)
class CliDir(CliType):
    pass


@dataclass(unsafe_hash=True)
class CliFile(CliType):
    pass


@dataclass(unsafe_hash=True)
class CliDict(CliType):
    key: CliType
    value: CliType


@dataclass(unsafe_hash=True)
class CliList(CliType):
    value: CliType


@dataclass(unsafe_hash=True)
class CliTuple(CliType):
    values: typing.List[CliType]

    @property
    def homogenous(self):
        # A tuple is homogenous if all types in the tuple are the same, aka the set of all types has length 1
        return len(set([type(x) for x in self.values])) == 1
